from .base import BaseNotificationChannel


class ConsoleChannel(BaseNotificationChannel):
    """Dummy channel that prints to the console."""

    def construct_message(self):
        """Stringify the notification kwargs."""
        return str(self.notification_kwargs)

    def notify(self, message):
        print(message)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2017-10-06 01:26
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0006_auto_20171005_0402'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='obj',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='notification',
            name='url',
            field=models.URLField(blank=True, null=True),
        ),
    ]



"""Channels backend"""

from __future__ import absolute_import, unicode_literals

import channels.layers
from asgiref.sync import async_to_sync

from .base import BaseBackend
from .. import default_settings as settings


class ChannelsBackend(BaseBackend):

    def run(self):
        channel_layer = channels.layers.get_channel_layer(
            settings.NOTIFICATIONS_QUEUE_NAME
        )

        for channel_alias in self.notification['channels']:
            async_to_sync(channel_layer.send)(
                settings.NOTIFICATIONS_QUEUE_NAME,
                {
                    'notification': self.notification,
                    'channel_alias': channel_alias, 'type': 'notify'
                }
            )



# Generated by Django 2.0.1 on 2018-01-28 08:27

from django.db import migrations
import notifications.fields


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0009_auto_20180112_0915'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='channels',
            field=notifications.fields.ListField(default=('console',), max_length=200),  # noqa
            preserve_default=False,
        ),
    ]



"""Channels consumers."""

import json
import logging

from channels.consumer import SyncConsumer
from channels.generic.websocket import AsyncWebsocketConsumer

from . import default_settings as settings
from notifications.backends.utils import _send_notification


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('django_notifs.backends.channels')


class DjangoNotifsConsumer(SyncConsumer):

    def notify(self, message):
        _send_notification(
            message['notification'], message['channel_alias'], logger
        )


class DjangoNotifsWebsocketConsumer(AsyncWebsocketConsumer):
    async def connect(self):
        self.room_group_name = self.scope['url_route']['kwargs'][
            settings.NOTIFICATIONS_WEBSOCKET_URL_PARAM
        ]

        # Join room group
        await self.channel_layer.group_add(
            self.room_group_name,
            self.channel_name
        )

        await self.accept()

    async def disconnect(self, close_code):
        """Leave the group."""
        await self.channel_layer.group_discard(
            self.room_group_name,
            self.channel_name
        )

    async def receive(self, text_data):
        """
        Receive message from the WebSocket

        for bi-directional communication)
        """
        text_data_json = json.loads(text_data)
        message = text_data_json['message']

        # Send message to group
        await self.channel_layer.group_send(
            self.room_group_name,
            {
                'type': settings.NOTIFICATIONS_WEBSOCKET_EVENT_NAME,
                'message': message
            }
        )


async def __websocket_message(self, event):
    """Receive messages from the group."""
    message = event['message']

    # Send message to WebSocket
    await self.send(message)


setattr(
    DjangoNotifsWebsocketConsumer,
    settings.NOTIFICATIONS_WEBSOCKET_EVENT_NAME, __websocket_message
)



"""Utilities and helper functions."""

import importlib

from . import NotificationError
from .models import Notification
from . import default_settings as settings


def notify(silent=False, **kwargs):
    """Helper method to send a notification."""
    notification = Notification(**kwargs)

    # Validate channels
    for channel_alias in notification.channels:
        _validate_channel_alias(channel_alias)

    # If it's a not a silent notification, save the notification
    if not silent:
        notification.save()

    # Send the notification asynchronously with celery
    notification_delivery_backend = _import_class_string(
        settings.NOTIFICATIONS_DELIVERY_BACKEND
    )
    notification_delivery_backend(notification).run()


def read(notify_id, recipient):
    """
    Helper method to read a notification.

    Raises NotificationError if the user doesn't have access
    to read the notification
    """
    notification = Notification.objects.get(id=notify_id)

    if recipient != notification.recipient:
        raise NotificationError('You cannot read this notification')

    notification.read()


def _validate_channel_alias(channel_alias):
    """
    Validates a channel alias against settings.NOTIFICATION_CHANNELS.

    returns the channel's path (i.e the path to the Channel's class)
    raises an AttributeError for invalid aliases
    """
    try:
        channel_path = settings.NOTIFICATIONS_CHANNELS[channel_alias]
    except KeyError:
        raise AttributeError(
            '"%s" is not a valid delivery channel alias. '
            'Check your applications settings for NOTIFICATIONS_CHANNELS'
            % channel_alias
        )

    return channel_path


def _import_class_string(path):
    """helper to import classes from string paths."""
    package, attr = path.rsplit('.', 1)

    return getattr(importlib.import_module(package), attr)



"""Notification channels for django-notifs."""

from json import dumps

from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer

from .base import BaseNotificationChannel
from .. import default_settings as settings


class WebSocketChannel(BaseNotificationChannel):
    """django-channels websocket channel"""

    destination_name = settings.NOTIFICATIONS_WEBSOCKET_URL_PARAM

    def _validate_notification_kwargs(self):
        notif_kwargs = self.notification_kwargs

        try:
            notif_kwargs['extra_data'][self.destination_name]
        except KeyError:
            raise KeyError(
                f'`extra_kwargs` must contain the `{self.destination_name}`'
                'key'
            )

    def construct_message(self):
        """Construct the message to be sent."""
        self._validate_notification_kwargs()

        extra_data = self.notification_kwargs['extra_data']

        return dumps(extra_data['message'])

    def notify(self, message):
        channel_layer = get_channel_layer()
        destination = self.notification_kwargs['extra_data'][
            self.destination_name
        ]

        async_to_sync(channel_layer.group_send)(
            destination,
            {
                'type': settings.NOTIFICATIONS_WEBSOCKET_EVENT_NAME,
                'message': message
            }
        )



# Generated by Django 3.1.7 on 2021-03-13 19:00

from django.db import migrations
import django_jsonfield_backport.models
import notifications.fields


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0001_squashed_0010_notification_channels'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='channels',
            field=notifications.fields.ListField(max_length=200),
        ),
        migrations.AlterField(
            model_name='notification',
            name='extra_data',
            field=django_jsonfield_backport.models.JSONField(default=dict),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-07-18 15:39
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='source',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),  # noqa
        ),
        migrations.AlterField(
            model_name='notification',
            name='source_display_name',
            field=models.CharField(max_length=150, null=True),
        ),
    ]



"Utility functions"

from ..utils import _import_class_string, _validate_channel_alias


def _send_notification(notification, channel_alias, logger):
    """Instantiates the notif-channel and calls it's 'notify' method."""
    # Validate channel alias
    channel_path = _validate_channel_alias(channel_alias)

    channel = _import_class_string(channel_path)(
        **notification, alias=channel_alias
    )

    message = channel.construct_message()
    channel.notify(message)
    logger.info(
        'Sent notification with the %s channel. kwargs: %s\n' %
        (channel_alias, notification)
    )



from .rq import RQBackend as RQ  # noqa
from .celery import CeleryBackend as Celery  # noqa
from .django_channels import ChannelsBackend as Channels  # noqa
from .synchronous import SynchronousBackend as Synchronous  # noqa



"""Synchronous backend"""

import logging

from .base import BaseBackend
from .utils import _send_notification


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('django_notifs.backends.synchronous')


class SynchronousBackend(BaseBackend):

    def run(self):
        for channel_alias in self.notification['channels']:
            _send_notification(self.notification, channel_alias, logger)



"""Celery backend"""

from __future__ import absolute_import, unicode_literals


from celery import shared_task
from celery.utils.log import get_task_logger

from .base import BaseBackend
from .utils import _send_notification
from .. import default_settings as settings


logger = get_task_logger(__name__)

autoretry_for = []
if settings.NOTIFICATIONS_RETRY:
    autoretry_for.append(Exception)


@shared_task(
    autoretry_for=autoretry_for,
    retry_backoff=settings.NOTIFICATIONS_RETRY_INTERVAL,
    max_retries=settings.NOTIFICATIONS_MAX_RETRIES
)
def send_notification(notification, channel_alias):
    """Send notification via a channel to celery."""
    _send_notification(notification, channel_alias, logger)


class CeleryBackend(BaseBackend):
    def run(self):
        for channel_alias in self.notification["channels"]:
            send_notification.apply_async(
                args=[self.notification, channel_alias],
                queue=settings.NOTIFICATIONS_QUEUE_NAME,
            )



# Generated by Django 2.0 on 2017-12-28 20:07

from django.db import migrations
import notifications.fields


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0007_auto_20171006_0126'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='extra_data',
            field=notifications.fields.JSONField(default={}),
        ),
    ]



"""Default settings for django-notifs project."""

from django.conf import settings


NOTIFICATIONS_DELIVERY_BACKEND = getattr(
    settings, 'NOTIFICATIONS_DELIVERY_BACKEND',
    'notifications.backends.Synchronous'
)

NOTIFICATIONS_QUEUE_NAME = getattr(
    settings, 'NOTIFICATIONS_QUEUE_NAME', 'django_notifs'
)

NOTIFICATIONS_RETRY = getattr(settings, 'NOTIFICATIONS_RETRY', False)

NOTIFICATIONS_RETRY_INTERVAL = getattr(
    settings, 'NOTIFICATIONS_RETRY_INTERVAL', 5
)

NOTIFICATIONS_MAX_RETRIES = getattr(settings, 'NOTIFICATIONS_MAX_RETRIES', 5)

NOTIFICATIONS_CHANNELS = getattr(settings, 'NOTIFICATIONS_CHANNELS', {})

NOTIFICATIONS_WEBSOCKET_EVENT_NAME = getattr(
    settings, 'NOTIFICATIONS_WEBSOCKET_EVENT_NAME', 'notifs_websocket_message'
)
NOTIFICATIONS_WEBSOCKET_URL_PARAM = getattr(
    settings, 'NOTIFICATIONS_WEBSOCKET_URL_PARAM', 'room_name'
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-07-19 11:38
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0003_auto_20170719_1137'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='notification',
            options={'ordering': ('-create_date',)},
        ),
    ]



import abc


class BaseNotificationChannel(metaclass=abc.ABCMeta):
    """Base channel for sending notifications."""

    def __init__(self, **kwargs):
        self.notification_kwargs = kwargs

    @abc.abstractmethod
    def construct_message(self):
        """Constructs a message from notification details."""
        raise NotImplementedError

    @abc.abstractmethod
    def notify(self, message):
        """Sends the notification."""
        raise NotImplementedError



"""Base backend."""

import abc


class BaseBackend(metaclass=abc.ABCMeta):

    def __init__(self, notification):
        self.notification = notification.to_json()

    @abc.abstractmethod
    def run(self):
        pass



"""Adpaters to send notifications through various meduiums."""

from django.conf import settings

from .base import BaseNotificationChannel

import requests


class PushNotificationChannel(BaseNotificationChannel):
    """Push notification channel."""

    def notify(self, message):
        """Send the notification."""
        kwargs = self.notification_kwargs
        # message = self.construct_message()

        # Send push notification(s)
        subject = '{} {} {}'.format(
            kwargs['source_display_name'], kwargs['action'],
            kwargs.get('short_description', '')
        )
        icon = kwargs.get('icon')
        token = kwargs.get('extra_data', {}).get('token', '')

        if token:
            requests.post(
                'https://fcm.googleapis.com/fcm/send',
                json={
                    'notification': {
                        'title': subject,
                        # 'body': message,
                        'click_action': kwargs.get('url', ''),
                        'icon': icon
                    },
                    'to': token
                },
                headers={'Authorization': 'key=' + settings.FCM_KEY}
            )



"""Django channels routes."""

from django.urls import path

from . import default_settings as settings
from .consumers import DjangoNotifsWebsocketConsumer

websocket_urlpatterns = [
    path(
        f'<{settings.NOTIFICATIONS_WEBSOCKET_URL_PARAM}>',
        DjangoNotifsWebsocketConsumer.as_asgi()
    )
]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-10-05 04:02
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0005_auto_20170814_1118'),
    ]

    operations = [
        migrations.RenameField(
            model_name='notification',
            old_name='recipent',
            new_name='recipient',
        ),
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import find_packages, setup


# Package meta-data.
NAME = 'django-notifs'
DESCRIPTION = 'Modular notifications for Django'
URL = 'https://github.com/danidee10/django-notifs'
EMAIL = 'osaetindaniel@gmail.com'
AUTHOR = 'Osaetin Daniel'
REQUIRES_PYTHON = '>=3.5.0'
VERSION = '3.0.0'

REQUIRED = [
    'django>=2.0', 'requests==2.25.1', 'django-jsonfield-backport==1.0.3'
]
TEST_REQUIRES = ['coverage>=5.4']
EXTRAS_REQUIRE = {
    'celery': ['celery>=4.1.0'],
    'rq': ['django-rq>=2.4.0'],
    'channels': ['channels>=3.0.3', 'channels-redis>=3.2.0']
}
EXCLUDE = ['notifs', 'tests', '*.tests', '*.tests.*', 'tests.*']

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=DESCRIPTION,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=EXCLUDE),
    install_requires=REQUIRED,
    test_requires=TEST_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    license='MIT'
)



"Notifications app config"

from django.apps import AppConfig


class NotificationsConfig(AppConfig):
    """App Config."""

    name = 'notifications'
    default_auto_field = 'django.db.models.AutoField'



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-06-24 00:20
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Notification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),  # noqa
                ('source_display_name', models.CharField(max_length=150)),
                ('action', models.CharField(max_length=50)),
                ('category', models.CharField(max_length=50)),
                ('obj', models.IntegerField()),
                ('url', models.URLField()),
                ('is_read', models.BooleanField(default=False)),
                ('create_date', models.DateTimeField(auto_now_add=True)),
                ('update_date', models.DateTimeField(auto_now=True)),
                ('short_description', models.CharField(max_length=100)),
                ('recipent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='notifications', to=settings.AUTH_USER_MODEL)),  # noqa
                ('source', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),  # noqa
            ],
        ),
    ]



"""Custom model fields."""

from django.db import models

from django_jsonfield_backport.models import JSONField  # noqa


class ListField(models.CharField):
    """Emulate django.contrib.postgres ArrayField."""

    def get_prep_value(self, value):
        """Convert the list/tuple to a comma separated string."""
        super(ListField, self).get_prep_value(value)

        return ','.join(value)

    def from_db_value(self, value, expression, connection):
        """Return the value from the database as a python list."""

        return value.split(',')



# Generated by Django 2.0.1 on 2018-01-12 09:15

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0008_notification_extra_data'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='recipient',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='notifications', to=settings.AUTH_USER_MODEL),  # noqa
        ),
    ]



from .console import ConsoleChannel  # noqa
from .base import BaseNotificationChannel  # noqa
from .channels_websocket import WebSocketChannel  # noqa
from .push_notification import PushNotificationChannel  # noqa



"""RQ Backend"""

import logging

import django_rq
from rq import Retry

from .base import BaseBackend
from .utils import _send_notification
from .. import default_settings as settings


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('django_notifs.backends.rq')

retry = None
if settings.NOTIFICATIONS_RETRY:
    retry = Retry(
        max=settings.NOTIFICATIONS_MAX_RETRIES,
        interval=settings.NOTIFICATIONS_RETRY_INTERVAL
    )


class RQBackend(BaseBackend):

    def run(self):
        for channel_alias in self.notification['channels']:
            queue = django_rq.get_queue(settings.NOTIFICATIONS_QUEUE_NAME)
            queue.enqueue(
                _send_notification, self.notification, channel_alias, logger,
                retry=retry
            )
